/**
 * ArffParser.java
 * @author Giorgio Corani (giorgio@idsia.ch)
 * 
 * Copyright:
 * Giorgio Corani, Marco Zaffalon
 *
 * IDSIA
 * Istituto Dalle Molle di Studi sull'Intelligenza Artificiale
 * Manno, Switzerland
 * www.idsia.ch
 * 
 *
 * The JNCC distribution is free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation (either version 2 of the License or, at your option, any later
 * version), provided that this notice and the name of the author appear in all 
 * copies. JNCC is distributed "as is", in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
 * more details.
 * You should have received a copy of the GNU General Public License
 * along with the JNCC distribution. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package jncc20;


import java.io.BufferedReader;
import java.io.File;
import java.io.LineNumberReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;



/**
 *Implementation of ARFF parser, used to parse training and testing files.
 *Remarks:
 * <p>
 * 1) Variables are supposed to be either nominal or numerical. Unlike Weka,
 * it does *not* manage variable of type String, or Date.
 * <p>
 * 2)It assumes the class of the problem to be named "class" in the Arff file and to be
 * declared as <em>last</em> variable in the header.
 */

class ArffParser {


	/**
	 * Initializes data members; than, scans the Arff file, checking the formal correctness 
	 * of variable declarations, and  the coherence of the data with the declarations; stores the information and the data
	 * loaded from file.
	 * In particular, it instantiates the data members FeatureNames, NumFlags(whether every feature is numerical or not), 
	 * CategoryNames(names of categories for each
	 * categorical featrue) and RawDataset (a matrix of double which contains the data as read from file, with missing values substitued by -9999, 
	 * and category names substituted by numerical indexes, and numerical values unchanged.)
	 * Moreover, reads the list of NonMar variables, which are then stored in NonMarFeatureNamesTraining and
	 * NonMarFeatureNamesTesting. 
	 */
	ArffParser (String UserSuppliedWorkingPath,
			String UserSuppliedArffName, String UserSuppliedValidationMethod) {

		if (!UserSuppliedArffName.endsWith(".arff")) {
			System.out.println("Data file " + UserSuppliedArffName
					+ " has different extension from the expected" + ".arff");
			System.exit(0);
		}
		validationMethod=UserSuppliedValidationMethod;

		try
		{
			File WorkingPathFile;
			WorkingPathFile = new File (UserSuppliedWorkingPath);
			workingPath=WorkingPathFile.getCanonicalPath();

			if (!(WorkingPathFile.getCanonicalPath().endsWith(System.getProperty("file.separator"))))
				workingPath +=System.getProperty("file.separator");

		}
		catch (IOException e)
		{
			System.err.println(e);
		}

		arffFileAddress = workingPath + UserSuppliedArffName;

		parseArffFile();

	};


	/**
	 * Reads the file NonMar.txt, containing the list of nonMar variables; if no
	 * file is found, all variables are assumed to be MAR.
	 * If the name of the variable is not preceeded by any token, the feature is supposed to be NonMar on
	 * both training and testing set; if it is preceeded by "training" ["testing"], then it is managed as NonMar on training
	 * [testing] only, and hence as Mar on testing [training].
	 * <p>
	 * Then, put the names of NonMar variables in TrainingNonMarFeatureNames and TestingNonMarFeatureNames.
	 */
	private void readNonMar() {

		BufferedReader bfr = null;
		String NonMarFile = workingPath + "NonMar.txt";
		nonMarFeatureNamesTraining = new ArrayList<String>();
		nonMarFeatureNamesTesting  = new ArrayList<String>();

		try {
			bfr = new LineNumberReader(new InputStreamReader(new FileInputStream(
					NonMarFile)));
			String line = "";
			try {
				String tmpStr;
				while ((line = bfr.readLine()) != null) {
					StringTokenizer st = new StringTokenizer(line);

					if (st.countTokens()>2)
					{
						System.out.println("Error in file "+NonMarFile);
						System.out.println("Maximum of two tokens per line is allowed");
						System.out.println("Check that every line  is formed either as either:\n");
						System.out.println("training[or testing] <name of the feature>");
						System.out.println("or:");
						System.out.println("<name of the feature>");
						System.out.println("In the latter case, the feature is assumed NonMar both on training and testing");
						System.exit(0);
					}
					tmpStr=st.nextToken();
					if  (tmpStr.compareToIgnoreCase("nonmar")==0) {
						if (((nonMarFeatureNamesTesting.size()==0 & nonMarFeatureNamesTraining.size()==0)) & ((line = bfr.readLine()) == null))
						{
							nonMarFeatureNamesTesting=featureNames;
							nonMarFeatureNamesTraining=featureNames;
							System.out.println("All features assumed to be NonMAR!");
							return;
						}
						else 
						{
							System.out.println("Possible inconsistency in NonMAR declarations:");
							System.out.println("I don't expect any further line in file if you use the token 'nonmar' ");
							System.out.println("Please setup properly file: "+NonMarFile);
							System.exit(0);
						}

					}


					if (tmpStr.equalsIgnoreCase("training"))
					{
						tmpStr=st.nextToken();
						if (nonMarFeatureNamesTraining.contains(tmpStr))
						{
							System.out.println("Feature "+tmpStr+" declared more than once as NonMar in training. Please fix file "+workingPath+ " NonMar.txt");
							System.exit(0);
						}
						else
							nonMarFeatureNamesTraining.add(tmpStr);
					}

					else if (tmpStr.equalsIgnoreCase("testing"))
					{
						tmpStr=st.nextToken();
						if (nonMarFeatureNamesTesting.contains(tmpStr))
						{
							System.out.println("Feature "+tmpStr+" declared more than once as NonMar in testing. Please fix file "+workingPath+ " NonMar.txt");
							System.exit(0);
						}
						else
							nonMarFeatureNamesTesting.add(tmpStr);
					}

					//if it is the feature name, no further tokens are allowed
					else if (st.countTokens()>0)
					{
						System.out.println("Error in file "+NonMarFile);
						System.out.println("\nCheck that every line  is formed either as either:\n");
						System.out.println("training[or testing] <name of the feature>");
						System.out.println("\nor:");
						System.out.println("<name of the feature>");
						System.out.println("\nIn the latter case, the feature is assumed NonMar both on training and testing");
						System.exit(0);
					}


					//NonMar on both training and testing set
					else
					{
						if (nonMarFeatureNamesTesting.contains(tmpStr))
						{
							System.out.println("Feature "+tmpStr+" declared more than once as NonMar in testing. Please fix file "+workingPath+ " NonMar.txt");
							System.exit(0);
						}

						if (nonMarFeatureNamesTraining.contains(tmpStr))
						{
							System.out.println("Feature "+tmpStr+" declared more than once as NonMar in training. Please fix file "+workingPath+ " NonMar.txt");
							System.exit(0);
						}
						nonMarFeatureNamesTesting.add(tmpStr);
						nonMarFeatureNamesTraining.add(tmpStr);
					}

				}
			} catch (Exception e) {
			}
		} 
		catch (FileNotFoundException exc) {

			//is perhaps the case wrong?
			File CurrentDir= new File (workingPath); 
			String[] fileNames = CurrentDir.list();
			for (String tmpString:fileNames)
			{
				if (tmpString.compareToIgnoreCase("NonMar.txt")==0)
				{
					System.out.println("Missing file NonMar.txt in directory "+ workingPath);
					System.out.println("Found file: "+tmpString);
					System.out.println("Please rename it as NonMar.txt if it contains declarations of NonMAR features for current experiment"); 
					System.out.println("Otherwise remove it, if has not to do with the experiment");
					System.exit(0);
				}
			}
			//we arrive here only if a file different only for the case has not been found; in this case, we assume 
			//the NonMAR file is deliberately missing
			System.out.println("\n\nFile "+ NonMarFile+" not found; continue assuming all features to be MAR");
		}
	}


	/**
	 * Scans the main Arff file.
	 */
	private void parseArffFile() {

		try {

			//System.out.println("\nParsing file: "+ArffFileAddress+"...");
			LineNumberReader lnr;
			String line = "";
			String currentToken = "";
			String[] tmpStrAr;
			boolean DataFound = false;
			int i, j;
			int FeatureIdx = 0;
			int tmpIndex = 0;

			// dynamic structures for managing data of unknown size
			featureNames = new ArrayList<String>();
			categoryNames = new ArrayList<String[]>();
			numFlags = new ArrayList<Boolean>();
			datasetName=null;

			lnr = new LineNumberReader(new InputStreamReader(new FileInputStream(arffFileAddress)));
			
			// begin scan file header
			while (DataFound == false) {
				line = lnr.readLine();
				// skip blank lines
				if (line.trim().equals(""))
					continue;
				StringTokenizer st = new StringTokenizer(line);
				currentToken = st.nextToken();

				// skip comments
				if (currentToken.charAt(0) == '%')
					continue;

				if (currentToken.equalsIgnoreCase("@RELATION")) {
					if (datasetName==null)
						datasetName = st.nextToken();
					else{
						System.out.println("Name of the relation is defined multiple times: "+datasetName+", "+st.nextToken());
						System.out.println("Fix the ARFF file by providing a single name for the relation.");
						System.exit(0);
					}

				} else if (currentToken.equalsIgnoreCase("@ATTRIBUTE"))
				{	
					currentToken = st.nextToken();

					if (! (featureNames.indexOf(currentToken)==-1))
					{
						System.out.println("File: "+arffFileAddress+"\nLine: "+lnr.getLineNumber()+
								"\nDuplicate attribute name "+currentToken);
						System.exit(0);
					}


					featureNames.add(currentToken);
					currentToken = st.nextToken();

					// we have a nominal feature? in this case, tokenizer
					// should
					// use commas as delimiters

					if (currentToken.charAt(0) == '{') {

						tmpIndex = line.indexOf('{');
						currentToken = line.substring(tmpIndex).trim();

						if (!(currentToken
								.charAt(currentToken.length() - 1) == '}')) {
							System.out.println("Line: "
									+ lnr.getLineNumber() + " of file:"
									+ arffFileAddress);
							System.out
							.println("List of values not enclosed into {} brackets");
							System.exit(0);
						}

						// remove brackets from token
						currentToken = currentToken.substring(1,
								(currentToken.length() - 1)).trim();
						tmpStrAr = (currentToken.split(",", -1));

						for (i = 0; i < tmpStrAr.length; i++) {
							tmpStrAr[i] = tmpStrAr[i].trim();
						}
						categoryNames.add(tmpStrAr);
						numFlags.add(false);
						FeatureIdx++;

					}

					else {

						// we have a numerical feature
						if (currentToken.equalsIgnoreCase("real")
								| currentToken.equalsIgnoreCase("numeric") 
								| currentToken.equalsIgnoreCase("integer")) {
							String[] Empty = { "empty" };
							categoryNames.add(Empty);
							numFlags.add(true);
							FeatureIdx++;
						} else {
							System.out.println("Line "
									+ lnr.getLineNumber() + " of file: "
									+ arffFileAddress);
							System.out
							.println("Non nominal features should be declared either as "
									+ "real or numeric");
							System.out
							.println("Or, if the feature is a nominal one, its values should be enclosed "
									+ " into { } brackets");

							System.exit(0);
						}

					}//==END NUMERICAL FEATURE
				}//==END "@ATTRIBUTE"

				// we do not accept tokens other than @attribute, @relation,
				// @data,
				else if (!(currentToken.equalsIgnoreCase("@DATA"))) {
					System.out.println("Wrong token in line "
							+ lnr.getLineNumber() + " of file " + arffFileAddress);
					System.exit(0);
				}

				else if (currentToken.equalsIgnoreCase("@DATA")) {

					//we have to remove the last feature from the feature set, as it is going to be the class.

					//Previously, however, we have to check that it is not numerical, otherwise the problem cannot be addressed.
					if (numFlags.get(numFlags.size()-1)){
						System.out.println("Last declared feature is numerical, and therefore cannot be used for a classification problem");
						System.exit(0);
					}

					int lastIdx=featureNames.size()-1;
					//now that the last feature is known to be a nominal one, we have to use it as a class and to drop it from the feature set.
					classNames = new ArrayList<String>();
					classNames= new ArrayList<String>(Arrays.asList(categoryNames.get(lastIdx)));
					featureNames.remove(lastIdx);
					numFlags.remove(lastIdx);
					categoryNames.remove(lastIdx);
					DataFound = true;
				}
			}// end (while DataFound==false)

			// At this point, the header has been entirely parsed
			// Let's load the name of NonMar variables and check if they are
			// consistent with the names
			// of the variables already parsed
			// Execution is halted if no correspondence is found.
			readNonMar();

			if (!(nonMarFeatureNamesTraining.isEmpty())) {
				for (String tmpStrn : nonMarFeatureNamesTraining) {
					if (featureNames.indexOf(tmpStrn) == -1) {
						System.out.println("\n\nNonMar variable " + tmpStrn.toUpperCase()+
								" declared in file: "+workingPath+"NonMar.txt"
								+ "\nbut NOT existent in data-file " + arffFileAddress);
						System.exit(0);
					}
				}
			}

			if (!(nonMarFeatureNamesTesting.isEmpty())) {
				for (String tmpStrn : nonMarFeatureNamesTesting) {
					if (featureNames.indexOf(tmpStrn) == -1) {
						System.out.println("\n\nNonMar variable " + tmpStrn.toUpperCase()+
								" declared in file: "+workingPath+"NonMar.txt"
								+ "\nbut NOT existent in data-file " + arffFileAddress);
						System.exit(0);
					}
				}
			}


			double[] CurrentRow;
			boolean catNameChecked;
			boolean classNameChecked;
			RawDataset = new ArrayList<double[]>();

			// ==TO DO ONLY IF CV
			if (validationMethod.equals("CV"))
			{
				rowsClassIdx = new ArrayList[classNames.size()];
				for (int ii=0;ii<classNames.size();ii++)
					rowsClassIdx[ii]=new ArrayList<Integer>();
			}

			// ===END ONLY CV

			// ==PARSE THE COMMA SEPARATED VALUES
			while ((line = lnr.readLine()) != null) {

				// current row: all the features and the class as last value
				CurrentRow = new double[(featureNames.size()) + 1];
				// skip blank lines
				if (line.trim().equals(""))
					continue;

				// parse data
				StringTokenizer st = new StringTokenizer(line, ",");
				currentToken = st.nextToken().trim();

				// skip commented line
				if (currentToken.charAt(0) == '%')
					continue;

				
				if (st.countTokens() != featureNames.size())
				{
					int tmpInt=featureNames.size()+1;
					int foundTokens=st.countTokens()+1;
					System.out.println("File: "+arffFileAddress+"\nLine: "+lnr.getLineNumber()
							+"\nExpected "+ tmpInt + " tokens \nFound " +
							foundTokens+" tokens" );
					System.exit(0);
				}

				// ==PARSE FEATURES VALUES IN THE CURRENT ROW
				for (i = 0; i < featureNames.size(); i++) {

					if (currentToken.equals("?"))
						CurrentRow[i] = -9999;

					// numerical feature
					else if (numFlags.get(i))
						try {
							CurrentRow[i] = Double.parseDouble(currentToken);
						}

					catch (NumberFormatException e) {
						System.out.println("Line " + lnr.getLineNumber()
								+ " of file " + arffFileAddress);
						System.out
						.println("Value "
								+ currentToken
								+ " incompatible with the numerical attribute "
								+ featureNames.get(i));
						System.exit(0);
					}

					// nominal feature
					else {
						catNameChecked = false;
						tmpStrAr = categoryNames.get(i);

						for (j = 0; j < tmpStrAr.length; j++) {
							if (currentToken.equalsIgnoreCase(tmpStrAr[j])) {
								catNameChecked = true;
								CurrentRow[i] = j;
								break;
							}
						} // end **for (j=0; j<tmpStrAr.length; j++)

						if (catNameChecked == false) {
							System.out.println("Line " + lnr.getLineNumber()
									+ " of file " + arffFileAddress);
							System.out
							.println("Value "
									+ currentToken
									+ " does not match any of the possible values"
									+ " of attribute "
									+ featureNames.get(i));
							System.exit(0);
						}

					}// end else (nominal feature)
					currentToken = st.nextToken().trim();
				}// end **for (int i=0; i<FeatureNames.size()(); i++)
				// ==FEATURES IN THE CURRENT ROW HAVE BEEN PARSED

				// == READ AND STORE CLASS VALUES
				classNameChecked = false;

				for (j = 0; j < classNames.size(); j++) {
					if (currentToken.equalsIgnoreCase(classNames.get(j))) {
						classNameChecked = true;
						CurrentRow[featureNames.size()] = j;
						break;
					}
				}
				if (classNameChecked == false) {
					System.out.println("Line " + lnr.getLineNumber()
							+ " of file " + arffFileAddress);
					if (currentToken.equalsIgnoreCase("?")){
						System.out.println("Missing value not allowed for the class variable");
						System.exit(0);
					}
					else
					{
						System.out.println("Value " + currentToken
								+ " does not match any of the possible values"
								+ " for the output class");
						System.exit(0);
					}
				}
				// ==CLASS PARSED

				RawDataset.add(CurrentRow);

				// ==ONLY FOR CV CASE!!
				if (validationMethod.equals("CV")) 
					rowsClassIdx[j].add(RawDataset.size()-1);

			}// end parsing data (while line != null)


			lnr.close();
			//System.out.println("\nFile: "+ArffFileAddress+"  parsed");
		}// end try

		catch (FileNotFoundException exc) {
			System.out.println("Missing File " + arffFileAddress);
			System.out.println("Please check path and file name!");
			System.exit(0);
		} catch (IOException ioexc) {
			System.out.println("Unexpected exception reading file " + arffFileAddress);
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Unexpected exception reading file " + arffFileAddress);
			System.exit(0);
		}

	}// end parseArffFile


	

	/**
	 * Parses the testing file, checking that all declarations are coherent with those already loaded from the training
	 * Arff file; if the classes are unknown, it reads only the instances, without looking for the classes.
	 * Data are stored in TestingSet: nominal features are simply stored, while 
	 * numerical features are discretized using the bins available from DiscretizationIntervals.
	 */

	void parseTestingArffFile(boolean UnknownClasses) {

		try {
			System.out.println("Parsing testing file: "+arffTestingFileAddress+"...");
			LineNumberReader lnr;
			String line = "";
			String currentToken = "";
			String[] tmpStrAr;
			boolean ClassFound = false;
			boolean DataFound = false;
			int i, j;
			int tmpIndex;

			//num of features found in testing set which match, both as for order and name, those of training set.
			//for the testing file to be accepted, it is necessary to have eventually foundFeatures=featuresNames.size()
			int foundFeatures=0;

			lnr = new LineNumberReader(new InputStreamReader(
					new FileInputStream(arffTestingFileAddress)));

			// begin scan file header
			while (!DataFound) {
				line = lnr.readLine();
				// skip blank lines
				if (line.trim().equals(""))
					continue;
				StringTokenizer st = new StringTokenizer(line);
				currentToken = st.nextToken();

				// skip comments
				if (currentToken.charAt(0) == '%')
					continue;

				if (currentToken.equalsIgnoreCase("@RELATION")) {
					if  (! st.nextToken().equalsIgnoreCase(datasetName) )
					{
						System.out.println ("File: " + arffTestingFileAddress + "\n" +
								"The relation name "+ currentToken + " does not match the relation name "
								+ datasetName + " indicated in file " + arffFileAddress);
						System.exit(0);
					}

				} else if (currentToken.equalsIgnoreCase("@ATTRIBUTE"))

				{
					if (ClassFound)
					{
						System.out.println("File: "+arffTestingFileAddress+"\nLine: "+lnr.getLineNumber()+
								"\nFeature "+st.nextToken()+" exceeds the features declared in training file.");
						System.exit(0);
					}
					currentToken = st.nextToken();


					//if this condition is verified, then we are going to parse the class
					if  (foundFeatures==featureNames.size())
					{
						tmpIndex = line.indexOf('{');
						currentToken = line.substring(tmpIndex).trim();

						if (!(currentToken
								.charAt(currentToken.length() - 1) == '}')) {
							System.out.println("Line: "
									+ lnr.getLineNumber() + " of file:"
									+ arffTestingFileAddress);
							System.out
							.println("List of values not enclosed into {} brackets");
							System.exit(0);
						}

						// remove brackets from token
						currentToken = currentToken.substring(1,
								(currentToken.length() - 1)).trim();
						tmpStrAr = (currentToken.split(",", -1));

						if (tmpStrAr.length != classNames.size())
						{
							System.out.println("Number of categories of last attribute (assumed to be the class) " +
									"does not match between files \n"+ arffFileAddress + " and " + arffTestingFileAddress);
							System.exit(0);
						}

						for (i = 0; i < tmpStrAr.length; i++) {
							tmpStrAr[i] = tmpStrAr[i].trim();
							if (! (tmpStrAr[i].equalsIgnoreCase(classNames.get(i))))
							{
								System.out.println ("Problem with categories of the last attribute (assumed to represent the class):" +
										" : \n"
										+"class" + tmpStrAr[i] + " declared in file \n " + 
										arffTestingFileAddress + "\n does not match class" + 
										classNames.get(i) + " declared in \n" + arffFileAddress );
								System.exit(0);
							}
						}
						ClassFound=true;
					}// end analysis of the class

					//if instead a feature is to be parsed, the following procedure has to be followed
					else
					{

						if  (featureNames.get(foundFeatures).equalsIgnoreCase(currentToken))
							foundFeatures+=1;
						else
						{//file rejected because variables supplied in the wrong order
							System.out.println("File: "+arffTestingFileAddress+"\nLine: " + lnr.getLineNumber()
									+"\nVariable " + currentToken
									+" found. \nExpected variable: "+ featureNames.get(foundFeatures));
							System.exit(0);
						}

						currentToken = st.nextToken();

						// we have a nominal feature? in this case, tokenizer
						// should
						// use commas as delimiters

						if (currentToken.charAt(0) == '{') {

							//note that if we want to use foundFeature as an index, we have to decrement it of an unit,
							//as array are indexed starting from zero
							if (numFlags.get(foundFeatures-1))
							{
								System.out.println("File: "+arffTestingFileAddress+"\nLine: "
										+lnr.getLineNumber()+"\n**Feature: "+featureNames.get(foundFeatures-1)
										+" declared as categorical.\nThis is incoherent with the declaration" +
										" for the same feature in file:\n"  +
										arffFileAddress);

								System.exit(0);
							}


							tmpIndex = line.indexOf('{');
							currentToken = line.substring(tmpIndex).trim();

							if (!(currentToken
									.charAt(currentToken.length() - 1) == '}')) {
								System.out.println("Line: "
										+ lnr.getLineNumber() + " of file:"
										+ arffTestingFileAddress);
								System.out
								.println("List of values not enclosed into {} brackets");
								System.exit(0);
							}

							// remove brackets from token
							currentToken = currentToken.substring(1,
									(currentToken.length() - 1)).trim();
							tmpStrAr = (currentToken.split(",", -1));

							if (tmpStrAr.length != categoryNames.get(foundFeatures-1).length)
							{
								System.out.println("Feature: "+ featureNames.get(foundFeatures-1)+
										"\n incoherent class names declarations between files \n"+
										arffFileAddress + " and " + arffTestingFileAddress); 
								System.exit(0);

							}


							for (i = 0; i < tmpStrAr.length; i++) {
								tmpStrAr[i] = tmpStrAr[i].trim();
								if (! (tmpStrAr[i].equalsIgnoreCase(categoryNames.get(foundFeatures-1)[i])))
								{
									System.out.println ("Feature " + featureNames.get(foundFeatures-1) + " : \n"
											+"Category " + tmpStrAr[i] + " declared in file \n " + 
											arffTestingFileAddress + "\n does not match category " + 
											categoryNames.get(foundFeatures-1)[i] 
											                                   + " declared in \n" + arffFileAddress );
									System.exit(0);
								}
							}
						}

						else {

							// we have a numerical feature

							if (! numFlags.get(foundFeatures-1))
							{

								System.out.println("File: "+arffTestingFileAddress+"\nLine: "
										+lnr.getLineNumber()+"\n**Feature: "+featureNames.get(foundFeatures-1)
										+"declared as numerical.\nThis is incoherent with the declaration"+
										"for the same feature in file:\n" + arffFileAddress);

								System.exit(0);
							}

							if (!(currentToken.equalsIgnoreCase("real")
									| currentToken.equalsIgnoreCase("numeric") 
									| currentToken.equalsIgnoreCase("integer"))) 
							{	
								System.out.println("Line "
										+ lnr.getLineNumber() + " of file: "
										+ arffTestingFileAddress);
								System.out
								.println("Non nominal features should be declared either as "
										+ "real or numeric");								
								System.exit(0);
							}

						}//==END numerical feature
					}
				}//==END (if @attribute)


				// we do not accept tokens other than @attribute, @relation,
				// @data,
				else if (!(currentToken.equalsIgnoreCase("@DATA"))) {
					System.out.println("Wrong token in line "
							+ lnr.getLineNumber() + " of file " + arffTestingFileAddress);
					System.exit(0);
				}

				else if (currentToken.equalsIgnoreCase("@DATA")) {
					if ((!ClassFound) & (!UnknownClasses)) {
						System.out.println("Problem in file " + arffTestingFileAddress);
						System.out
						.println("Missing declaration of output class before the '@DATA' tag");
						System.exit(0);
					}
					DataFound = true;
				}
			}// end (while DataFound==false)
			// At this point, the header has been entirely parsed.

			//We now know that all the variables in the testing file are present and 
			//coherently ordered with those in the training file. 
			//We should still check that all the training features are
			//present in the testing file.

			if (foundFeatures != numFlags.size())
			{
				System.out.println(numFlags.size() + " features  in " + arffFileAddress);
				System.out.println(foundFeatures + " features in " + arffTestingFileAddress);
				System.out.println("\n**List of missing feature in testing file:");
				for (int kk=foundFeatures; kk<numFlags.size(); kk++)
				{
					System.out.println("\n"+featureNames.get(kk));
				}		
				System.exit(0);

			}


			//==END HEADER PARSING AND CHECKING


			//===PARSE THE DATA 
			int[] CurrentRow;
			String[] CurrentRawRow;
			boolean catNameChecked;
			boolean classNameChecked;

			//I approximately suppose that it will have more or less the same number of data than 
			//the training
			testingSet = new ArrayList<int[]>(RawDataset.size());
			rawTestingSet = new ArrayList<String[]>(RawDataset.size());
			int TotalVariables;
			int UsedVariables;

			if (UnknownClasses)
			{
				UsedVariables=usedFeatures.size();
				TotalVariables=featureNames.size();
			}
			else
			{
				UsedVariables=(usedFeatures.size())+1;
				TotalVariables=featureNames.size()+1;
			}



			// ==PARSE THE COMMA SEPARATED VALUES
			while ((line = lnr.readLine()) != null) {


				CurrentRow = new int[UsedVariables];
				CurrentRawRow = new String[TotalVariables];

				// skip blank lines
				if (line.trim().equals(""))
					continue;

				// parse data
				StringTokenizer st = new StringTokenizer(line, ",");
				currentToken = st.nextToken().trim();

				// skip commented line
				if (currentToken.charAt(0) == '%')
					continue;

				if (st.countTokens() != (TotalVariables-1))
				{
					int foundTokens=st.countTokens()+1;
					System.out.println("File: "+arffTestingFileAddress+"\nLine: "+lnr.getLineNumber()
							+"\nExpected "+ TotalVariables + " tokens \nFound " +
							foundTokens+" tokens" );
					System.exit(0);
				}

				int counter=0;
				// ==PARSE FEATURES VALUES IN THE CURRENT ROW
				for (i = 0; i < featureNames.size(); i++) {

					if (usedFeatures.indexOf(i)<0)
					{
						currentToken = st.nextToken().trim();
						continue;
					}


					CurrentRawRow[counter]=currentToken;

					if (currentToken.equals("?"))
						CurrentRow[counter] = -9999;

					// numerical feature
					else if (numFlags.get(i))
						try {
							CurrentRow[counter]=getDiscretizationIdx(Double.parseDouble(currentToken),i);
						}

					catch (NumberFormatException e) {
						System.out.println("Line " + lnr.getLineNumber()
								+ " of file " + arffTestingFileAddress);
						System.out
						.println("Value "
								+ currentToken
								+ " incompatible with the numerical attribute "
								+ featureNames.get(i));
						System.exit(0);
					}

					// nominal feature
					else {
						catNameChecked = false;
						for (j = 0; j < categoryNames.get(i).length; j++) {
							if (currentToken.equalsIgnoreCase(categoryNames.get(i)[j])) {
								catNameChecked = true;
								CurrentRow[counter] = j;
								break;
							}
						} // end **for (j=0; j<tmpStrAr.length; j++)

						if (catNameChecked == false) {
							System.out.println("Line " + lnr.getLineNumber()
									+ " of file " + arffTestingFileAddress);
							System.out
							.println("Value "
									+ currentToken
									+ " does not match any of the possible values"
									+ " of attribute "
									+ featureNames.get(i));
							System.exit(0);
						}

					}// end else (nominal feature)


					if ((!UnknownClasses)| (i < featureNames.size()-1))
						currentToken = st.nextToken().trim();

					counter++;
				}// end **for (int i=0; i<FeatureNames.size()(); i++)
				// ==FEATURES IN THE CURRENT ROW HAVE BEEN PARSED

				if (!UnknownClasses)
				{
					// == READ AND STORE CLASS VALUES
					classNameChecked = false;

					for (j = 0; j < classNames.size(); j++) {
						if (currentToken.equalsIgnoreCase(classNames.get(j))) {
							classNameChecked = true;
							CurrentRow[UsedVariables-1] = j;
							break;
						}
					}
					if (classNameChecked == false) {
						System.out.println("Line " + lnr.getLineNumber()
								+ " of file " + arffTestingFileAddress);
						System.out.println("Value " + currentToken
								+ " does not match any of the possible values"
								+ " for the output class");
						System.exit(0);
					}
					// ==CLASS PARSED
				}

				testingSet.add(CurrentRow);
				rawTestingSet.add(CurrentRawRow);
			}// end parsing data (while line != null)

			lnr.close();

		}// end try

		catch (FileNotFoundException exc) {
			System.out.println("Missing File " + arffTestingFileAddress);
			System.exit(0);
		} catch (IOException ioexc) {
			System.out.println("Unexpected exception reading file " + arffTestingFileAddress);
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Unexpected exception reading file " + arffTestingFileAddress);
			System.exit(0);
		}
	}//end parseTestingFile




	/**Return the bin in which a numerical value of a given feature falls.
	 */
	private int getDiscretizationIdx(Double currentValue,
			int FeatureIdx) {

		int NextIdx=1;
		int PreviousIdx=0;
		int NumCutPoints=discretizationIntervals[FeatureIdx].length;

		if (currentValue<discretizationIntervals[FeatureIdx][0])
			return 0;

		if (currentValue>discretizationIntervals[FeatureIdx][NumCutPoints-1])
			return NumCutPoints;

		//on testing data, it may happen that a datum is actually equal to a cutpoint
		int MatchingCutPoint=Arrays.binarySearch(discretizationIntervals[FeatureIdx], currentValue);

		if (MatchingCutPoint < 0)
		{
			//we have an intermediate value
			while (! (currentValue<discretizationIntervals[FeatureIdx][NextIdx]) &
					(currentValue>discretizationIntervals[FeatureIdx][PreviousIdx]) )
			{
				PreviousIdx=NextIdx++;
			}
			return 	NextIdx;
		}
		else
			return MatchingCutPoint;
	}


	//==GETTER AND SETTER METHODS
	ArrayList<String[]> getCategoryNames() {
		return categoryNames;
	}

	ArrayList<String >getClassNames() {
		return classNames;
	}


	String getDatasetName() {
		return datasetName;
	}


	ArrayList<String> getFeatureNames() {
		return featureNames;
	}


	ArrayList<String> getNonMarFeatureNamesTesting() {
		return nonMarFeatureNamesTesting;
	}


	ArrayList<String> getNonMarFeatureNamesTraining() {
		return nonMarFeatureNamesTraining;
	}

	ArrayList<Boolean> getNumFlags() {
		return numFlags;
	}


	ArrayList<double[]> getRawDataset() {
		return RawDataset;
	}


	ArrayList<String[]> getRawTestingSet() {
		return rawTestingSet;
	}


	ArrayList<Integer>[] getRowsClassIdx() {
		return rowsClassIdx;
	}

	ArrayList<int[]> getTestingSet() {
		return testingSet;
	}

	void setUsedFeatures(ArrayList<Integer> suppliedUsedFeatures) {
		usedFeatures = suppliedUsedFeatures;
	}

	void setDiscretizationIntervals(double[][] suppliedDiscretizationIntervals) {
		discretizationIntervals = suppliedDiscretizationIntervals;
	}


	void setNotUsedFeatures(ArrayList<Integer> suppliedNotUsedFeatures) {
		notUsedFeatures = suppliedNotUsedFeatures;
	}

	void setArffTestingFileAddress(String suppliedArffTestingFileAddress) {
		arffTestingFileAddress = suppliedArffTestingFileAddress;
	}

	// =====DATA MEMBERS=====

	/**
	 * Dataset Name as read from the field "@relation" in the Arff file
	 */
	private String datasetName;

	/**
	 * Names of  NonMar features in training
	 */
	private ArrayList<String> nonMarFeatureNamesTraining;

	/**
	 * Names of  NonMar features in testing
	 */
	private ArrayList<String> nonMarFeatureNamesTesting;

	/** Names of input features */
	private ArrayList<String> featureNames;

	/** Flags array, regarding wheter Features are numerical (1) or not (0)*/
	private ArrayList<Boolean> numFlags;


	/**
	 * Copy of the data read from Arff file (having hence -9999 as marker for
	 * missing data), and category names substituted by the corresponding
	 * indexes.)
	 */
	private ArrayList<double[]> RawDataset;

	/**
	 * Raw testing set exactly as read from file.
	 * Used when a class-less file is used, and we want later
	 * to dump to file the original instances followed by the last column
	 * containing the classification.
	 * Being declared as String[][], it can hosts number as well as
	 * categories.
	 */
	private ArrayList<String[]> rawTestingSet;

	/**
	 * Matrix of String with rows of different lenght, as different features
	 * (each row of the matrix corresponds to a different feature) can have
	 * different numbers of categories.
	 * <p>
	 */
	private ArrayList<String[]> categoryNames;

	/**
	 * Testing set, as accessed by the classifier: numerical
	 * variables discretized, category names substituted by their
	 * indexes, missing data marked as -9999, classes substituted with indexes.
	 */
	private ArrayList<int[]> testingSet;


	/**
	 * Matrix with rows of different length; stores the bin ranges for numerical
	 * features
	 */
	private double[][] discretizationIntervals;

	/**Indexes of used features (i.e., categorical features and numerical features discretized into several bins)
	 */
	private ArrayList<Integer> usedFeatures;

	/**
	 * Indexes of features that are not used (because discretized into a single bin)
	 */
	private ArrayList<Integer> notUsedFeatures;

	/**
	 * Indexes of the rows, in RawDataset, which have the same output class.
	 * For instance, the first row collects the indexes of all the rows in RawDataset
	 * having output
	 * class c1, and so on.
	 */
	private ArrayList<Integer>[] rowsClassIdx;

	/**
	 * Path where the files for the given experiment (Arff files, NonMar.txt) reside, and where the output will be saved
	 */
	private String workingPath;

	/** Absolute Path of the main Arff file */
	private String arffFileAddress;

	/** Name of the testing Arff file */
	private String arffTestingFileAddress;


	/** Set either to "CV" or to the name of the testing Arff file */
	private String validationMethod;

	/**
	 * Names of the output class.
	 */
	private ArrayList<String> classNames;



}
