/**
 * jncc.java
 * @author Giorgio Corani (giorgio@idsia.ch)
 * 
 * Copyright:
 * Giorgio Corani, Marco Zaffalon
 *
 * IDSIA
 * Istituto Dalle Molle di Studi sull'Intelligenza Artificiale
 * Manno, Switzerland
 * www.idsia.ch
 *
 * JNCC is free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation (either version 2 of the License or, at your option, any later
 * version), provided that this notice and the name of the author appear in all 
 * copies. JNCC is distributed "as is", in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
 * more details.
 * You should have received a copy of the GNU General Public License
 * along with the JNCC distribution. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jncc20;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Random;
import java.util.StringTokenizer;
import java.lang.management.ManagementFactory;
import com.sun.management.OperatingSystemMXBean;



/**
 *Main class of the project.
 *It loads data from the file specified by the user (using ArffParser objects); then,
 *it trains and validates NBC and NCC.
 *Class jncc implementes three kinds of experiments:
 * 1) 10 runs of <em>stratified</em> 10-folds cross-validation; 
 * 2) validation via testing file (single training/testing experiment)
 * 3)testing file with unknown classes.
 *In the first two cases, accuracy stats are reported to file (via ResultsReporter objects),
 *as the true classes are known.
 *In the last case, NCC predictions only are reported to file, as the true classes are unknown.
 *Numerical features are discretized via MDL-entropy-based supervised discretization (using MdlDiscretizer 
 *objects. Note that discretization intervals
 *are always computed on the training set, and then applied unchanged on the testing set.
 */

public class Jncc {

	/**
	 * Initializes the necessary data members, scans the main Arff file and then
	 * instantiates the data members FeatureNames, 
	 * NumFlags, CategoryNames and RawDataset.
	 */
	Jncc(String UserSuppliedWorkingPath,
			String UserSuppliedArffName, String UserSuppliedValidationName) {

		if (!UserSuppliedArffName.endsWith(".arff")) {
			System.out.println("Data file " + UserSuppliedArffName
					+ " has different extension from the expected" + ".arff");
			System.exit(0);
		}


		try
		{
			File WorkingPathFile;
			WorkingPathFile = new File (UserSuppliedWorkingPath);
			workPath=WorkingPathFile.getCanonicalPath();

			if (!(WorkingPathFile.getCanonicalPath().endsWith(System.getProperty("file.separator"))))
				workPath +=System.getProperty("file.separator");

		}
		catch (IOException e)
		{
			System.err.println(e);
		}

		if (UserSuppliedValidationName.equalsIgnoreCase("cv")) {
			//setting for 10-folds-cross-validation


			numCvFolds = 10;
			numCrossVRuns = 10;
			foldsSize=new int[numCvFolds];
			arffTestingFile="none";
			arffTestingFileAddress = "none";
			validationMethod = "CV";
		} 
		else if (!UserSuppliedValidationName.endsWith("arff")) {
			System.out.println("Third arguments should be either 'cv'");
			System.out.println("or a testing file with .arff extension");
			System.exit(0);
		}
		// regular Arff filename supplied for testing
		else {
			numCvFolds = 0;
			arffTestingFile = UserSuppliedValidationName;
			arffTestingFileAddress=workPath +arffTestingFile;
			validationMethod = "TestingFile";
		}


		//scan ARFF file and get information
		aParser = new ArffParser(workPath,UserSuppliedArffName,validationMethod);

		//DatasetName=AParser.getDatasetName();
		nonMarFeatureNamesTraining=aParser.getNonMarFeatureNamesTraining();
		nonMarFeatureNamesTesting=aParser.getNonMarFeatureNamesTesting();
		featNames=aParser.getFeatureNames();
		numFlags=aParser.getNumFlags();
		rawDataset=aParser.getRawDataset();
		categoryNames=aParser.getCategoryNames();
		rowsClassIdx=aParser.getRowsClassIdx();
		classesNames=aParser.getClassNames();
		discretLog=new int[featNames.size()];
	};

	/**
	 * Arguments of the main: (1) the working path; (2) the name of the main ArrfFile;
	 * (3) "cv" or the name of the testing ArffFile;
	 * (4)[OPTIONAL] "unknownClasses", in case the
	 * actual classes of the testing set are unknown.
	 */
	public static void main(String[] args) {

		//time tracking
		OperatingSystemMXBean mxbean= (OperatingSystemMXBean) ManagementFactory.getOperatingSystemMXBean() ;
		double previousCputime; double elapsedCputime; 
		previousCputime=mxbean.getProcessCpuTime();



		if ((args.length != 3)& (args.length != 4))
		{
			System.out.println(args.length + " program arguments are supplied");
			System.out
			.println("3 program arguments should be specified by the user");
			System.exit(0);
		}

		if  (args.length == 4)
		{
			if (args[3].compareToIgnoreCase("unknownClasses")!= 0)
			{
				System.out.println("Fourth argument, if specified, can be only 'UnknownClasses'");
				System.exit(0);

			}
		}


		//we call dm the jncc instance. dm means DataManager, which is what
		//this class actually does.
		Jncc dm = new Jncc(args[0], args[1], args[2]);

		//cross-validation
		if (dm.validationMethod.equals("CV"))
			dm.validateViaCV(args);

		// testing file, whose classes are known
		else if (args.length != 4) 
			dm.validateViaTestingFile(args[2]);

		//testing file, whose classes are unknown
		else
			dm.validateViaTestingFileUnknownClasses ();

		elapsedCputime=(double) ((mxbean.getProcessCpuTime()-previousCputime)/Math.pow(10,9));	
		System.out.println("Elapsed Cpu time: "+ elapsedCputime +"seconds");
	}

	/**
	 *Validates NBc and NCC via 10 runs of 10-folds cross-validation.
	 *Reports to file the relevant accuracy measures.
	 */
	private void validateViaCV(String[] args)

	{
		try{

			//==various initalizations, and console messages
			System.out.println("Validating via cross-validation");
			//args[1] contains the name of the main ARFF file
			resFile = "Results-CV-"+args[1];
			int tmpidx=resFile.lastIndexOf(".");
			resFile=resFile.substring(0,tmpidx);
			resFile =workPath+resFile+".txt";
			predsFile = "Predictions-CV-"+args[1];
			predsFile=predsFile.substring(0,predsFile.lastIndexOf("."));
			predsFile=workPath+predsFile+".csv";
			//remove a possible copy of the prediction file referring to a previous run
			BufferedWriter tmpBfWrt=new BufferedWriter(new FileWriter(predsFile,false));
			tmpBfWrt.close();
			//==



			//here starts the computation
			for (int i = 0; i < numCrossVRuns; i++) {


				int iLog=i+1;
				System.out.println("\n\n==Run "+iLog+"/"+numCrossVRuns+" of cross-validation");


				drawCVindexes();

				//==LOOP IMPLEMENTING 10-FOLD CV



				//j indexes the current fold used as testing

				for (int j = 0; j <numCvFolds; j++) {


					int jLog=j+1;
					System.out.println("Fold n. "+jLog+"/"+numCvFolds);

					//==BUILD TRAINING AND TESTING SET
					//collect RawData, i.e. numerical features have to be discretized
					ArrayList<double[]> RawTrainingData=new ArrayList<double[]>(rawDataset.size()-foldsSize[j]);
					ArrayList<double[]> RawTestingData=new ArrayList<double[]>(foldsSize[j]);
					for (int k=0; k<rawDataset.size(); k++)
					{
						if (cvFoldsIdx[k] != j)
							RawTrainingData.add(rawDataset.get(k).clone());

						else
							RawTestingData.add(rawDataset.get(k).clone());
					}
					discretizeNumFeaturesOnTrainingData(RawTrainingData);
					trainingSet=new ArrayList<int[]>(RawTrainingData.size());
					testingSet=new ArrayList<int[]>(RawTestingData.size());

					/*copy nominal features, and the numerical ones as discretized, 
			 discarding features discretized into a unique bin */
					prepareDataSetFromRawData(RawTrainingData, trainingSet);
					prepareDataSetFromRawData(RawTestingData, testingSet);

					//==TRAINING AND TESTING SET ARE NOW READY

					//NBC
					NaiveBayes NBC = new NaiveBayes(trainingSet, usedFeaturesNames, classesNames, 
							numClassForEachUsedFeature);
					NBC.classifyInstances(testingSet);

					//NCC
					NaiveCredalClassifier NCC= new NaiveCredalClassifier(trainingSet, usedFeaturesNames, classesNames, 
							numClassForEachUsedFeature, nonMarInCurrentTrainingDataset, nonMarInCurrentTestingDataset, 
							numClassesNonMarTesting);
					NCC.classifyInstances(testingSet);	

					predictionsToFileNbcNcc (jLog, NBC.getPredictedInstances(), NCC.getCredalPredictedInstances());
				}//END ==10 FOLDS			
			}//END ==10-CV-EXPERIMENTS


			//parse predictions, compute stats and write them to file
			ResultsReporter reporter= new ResultsReporter(discretLog, resFile, featNames, classesNames, 
					predsFile, workPath, numCvFolds, numCrossVRuns);
			reporter.writeDiscretizationLog(numCrossVRuns*numCvFolds);
			reporter.analyzePredictionsFileNbcNcc();
			reporter.writeStatsToFile();

			System.out.println("\n\nCLASSIFICATION SUCCESSFULLY PERFORMED;\nResults written to file:\n"+resFile);
		}
		catch (IOException e)
		{
			System.out.println("Problem in executing JNCC; check you have writing permission on the working directory.");
			System.exit(0);
		}
	}

	/**Validates  NBC and  NCC via testing file.
	 * Reports to file the relevant accuracy measures.
	 */
	private void validateViaTestingFile (String TestingFile)
	{

		//various init. and console messages
		System.out.println("Validating via testing file");
		resFile = "Results-"+arffTestingFile;
		int tmpidx=resFile.lastIndexOf(".");
		resFile=resFile.substring(0,tmpidx);
		resFile =workPath+resFile+".txt";
		predsFile = "Predictions-Testing-"+TestingFile;
		predsFile=predsFile.substring(0,predsFile.lastIndexOf("."));
		predsFile=workPath+predsFile+".csv";
		//==


		//computations start here

		//RawDataset contains in this case the training data
		discretizeNumFeaturesOnTrainingData(rawDataset);


		/*copy nominal features, and the numerical ones as discretized, 
 discarding features discretized into a unique bin */
		trainingSet=new ArrayList<int[]>(rawDataset.size());
		prepareDataSetFromRawData(rawDataset, trainingSet);

		//==reads the testign file, and fills the TestingFile matrix,
		//by using the available Discretization Intervals. (variables
		//discretized into a unique bin are discarded)
		aParser.setUsedFeatures(usedFeatures); 
		aParser.setNotUsedFeatures(notUsedFeatures);
		aParser.setDiscretizationIntervals(discretizationIntervals);
		aParser.setArffTestingFileAddress(arffTestingFileAddress);
		aParser.parseTestingArffFile(false);
		testingSet=aParser.getTestingSet();
		rawTestingSet=aParser.getRawTestingSet();

		NaiveBayes NBC= new NaiveBayes(trainingSet, usedFeaturesNames, classesNames, 
				numClassForEachUsedFeature);
		NBC.classifyInstances(testingSet);

		NaiveCredalClassifier Ncc= new NaiveCredalClassifier(trainingSet, usedFeaturesNames, classesNames, 
				numClassForEachUsedFeature, nonMarInCurrentTrainingDataset, nonMarInCurrentTestingDataset, 
				numClassesNonMarTesting);				
		Ncc.classifyInstances(testingSet);
		predictionsToFileNbcNcc (0,NBC.getPredictedInstances(), Ncc.getCredalPredictedInstances());


		ResultsReporter reporter= new ResultsReporter(discretLog, resFile, featNames, classesNames,
				predsFile, workPath, numCvFolds, numCrossVRuns);
		reporter.writeDiscretizationLog(1);
		reporter.analyzePredictionsFileNbcNcc();
		reporter.writeStatsToFile();
		System.out.println("CLASSIFICATION SUCCESSFULLY PERFORMED;\nResults written to file:\n"+resFile);	
	}

	/**
	 * Dumps to file the predictions issued by both NBC and NCC on testing set(s).
	 * It produces a file which contains the features (apart from the ones discretized
	 * into a unique bin, which can change between different runs of CV), the actual class, the NCC classification
	 * (i.e., a number of columns equal to the number of classes containing either the outputted class, or 6666 to mean that
	 * not all classes have been outputted by NCC), and finally the NBC prediction.
	 */
	private  void predictionsToFileNbcNcc(int NumFold, int[]NBCPredictions, int[][] CredalPredictions)
	{
		int NumClasses=classesNames.size();
		try 
		{
			BufferedWriter out=new BufferedWriter(new FileWriter(predsFile,true));

			for (int k=0; k<testingSet.size(); k++)
			{
				//fold number 
				out.write(NumFold+"\t");
				//features and actual class output
				for (int jj=0; jj<testingSet.get(0).length; jj++)
					out.write(testingSet.get(k)[jj]+"\t");

				for (int tmpint : CredalPredictions[k])
					out.write(tmpint+"\t");

				for (int tmpint=0; tmpint<NumClasses-CredalPredictions[k].length; tmpint++ )
					out.write("6666"+"\t");

				out.write(NBCPredictions[k]+"\n");
			}
			out.close();
		}
		catch (IOException e)
		{
			System.out.println("\nProblem in writing CD predictions to file PredFile\n");
		}
	}



	/**
	 *Learns NCC; classifies the instances of the testing file via NCC, and writes the classifications to file.
	 */
	private void validateViaTestingFileUnknownClasses ()
	{
		//==various initalizations, and console messages
		System.out.println("Using testing file with unknown classes");
		resFile = "Results-"+arffTestingFile;
		int tmpidx=resFile.lastIndexOf(".");
		resFile=resFile.substring(0,tmpidx);
		resFile =workPath+resFile+".txt";
		//==


		//The computation starts here
		//RawDataset contains in this case the training data
		discretizeNumFeaturesOnTrainingData(rawDataset);

		/*copy nominal features, and the numerical ones as discretized, 
		 discarding features discretized into a unique bin */
		trainingSet=new ArrayList<int[]>(rawDataset.size());
		prepareDataSetFromRawData(rawDataset, trainingSet);

		//==reads the testign file, and fills the TestingFile matrix,
		//by using the available Discretization Intervals. (variables
		//discretized into a unique bin are discarded)
		//flag true specifies that we don't know the actual classes.
		aParser.setUsedFeatures(usedFeatures); 
		aParser.setNotUsedFeatures(notUsedFeatures);
		aParser.setDiscretizationIntervals(discretizationIntervals);
		aParser.setArffTestingFileAddress(arffTestingFileAddress);
		aParser.parseTestingArffFile(true);
		testingSet=aParser.getTestingSet();
		rawTestingSet=aParser.getRawTestingSet();		

		NaiveCredalClassifier NCC= new NaiveCredalClassifier(trainingSet, usedFeaturesNames, classesNames, 
				numClassForEachUsedFeature, nonMarInCurrentTrainingDataset, nonMarInCurrentTestingDataset, 
				numClassesNonMarTesting);				

		NCC.classifyInstances(testingSet);	
		ResultsReporter reporter= new ResultsReporter(discretLog, resFile, featNames, classesNames, 
				predsFile, workPath, numCvFolds, numCrossVRuns);
		reporter.writeDiscretizationLog(numCrossVRuns*numCvFolds);
		NCC.instancesAndPredictionsToFileUnknownClasses(workPath, arffTestingFile, rawTestingSet);
		System.out.println("CLASSIFICATION SUCCESSFULLY PERFORMED;\nResults written to file:\n"+resFile);
	}


	/**
	 * Draws <em>stratified</em>  folders for cross-validation, instantiating CvFoldsIdx.
	 */
	private void drawCVindexes() {

		cvFoldsIdx = new int[rawDataset.size()];

		/*
		 * (N x 1) array. Filled with the indexes of the rows, sorted according
		 * to the class, and previously randomized class-by-class
		 */
		ArrayList<Integer> IdxArray = new ArrayList<Integer>(rawDataset.size());
		ArrayList<Integer> bufferArrayList;
		int i, j;
		/*
		 * note that the random seed is initialized every time the function is
		 * called, to avoid generate the same random sequences over the
		 * different runs
		 */
		Random RandGenerator = new Random();
		int currRandIdx; 

		// now let's randomize them class by class
		for (i = 0; i < rowsClassIdx.length; i++) {
			bufferArrayList = new ArrayList<Integer>();

			for (int kk = 0; kk < rowsClassIdx[i].size(); kk++) {
				bufferArrayList.add(rowsClassIdx[i].get(kk));
			}

			while (!bufferArrayList.isEmpty()) {
				// choose randomly an index between 0 and (number of elements-1)
				// remember the nextInt generates Int up to (argument - 1)
				currRandIdx = RandGenerator.nextInt(bufferArrayList.size());
				IdxArray.add(bufferArrayList.get(currRandIdx));
				bufferArrayList.remove(currRandIdx);
			}// == end while
		}// end== for (int i=0; i<RawDatasetClassIdx.lenght; i++)

		/*
		 * At this point, IdxArray contain the indexes randomized class by
		 * class, and the array presents all the randomized indexes of class 1,
		 * then those of class 2, etc. Let's fill the CVFolderIdx
		 */

		int start = 0;
		int counter = 0;
		Arrays.fill(foldsSize,0);


		while (counter < IdxArray.size()) {
			j = start;

			while (j < IdxArray.size()) {
				cvFoldsIdx[IdxArray.get(j)] = start;
				j += numCvFolds;
				counter++;
				foldsSize[start]++;
			}// ==end of while

			start++;
		}// ==end while (counter<IdxArray.length)
	}// end== drawCVindexes


	/**Take a raw set of data (undiscretized features) and put them into 
	 * a dataset to be accessed by classifiers; categorical variables are copied unchanged, while
	 * numerical variables are converted to categorical according to DiscretizationIntervals;
	 * numerical variables discretized into a unique bin (and hence listed 
	 * in NonUsedFeatures) are discarded.
	 */
	private void prepareDataSetFromRawData(ArrayList<double[]> SourceData, ArrayList<int[]> DestinationData){

		int i,j;

		int MaxColumns=SourceData.get(0).length;
		//TrainingSet= new int[SourceData.length][];

		int[] tmpArr;
		int rowLength=usedFeatures.size()+1;

		for (i = 0; i < SourceData.size(); i++) {
			tmpArr=new int[rowLength];
			for (j=0; j<rowLength-1; j++)
			{
				if (! numFlags.get(usedFeatures.get(j)))
					tmpArr[j]=(int) SourceData.get(i)[usedFeatures.get(j)];

				else
					//numerical feature to discretize
				{
					if (SourceData.get(i)[usedFeatures.get(j)]!= -9999)
						tmpArr[j]=getDiscretizationIdx(SourceData.get(i)[usedFeatures.get(j)], usedFeatures.get(j));
					else
						tmpArr[j]=-9999;
				} 
			}//==Features have been copied

			//class copy
			tmpArr[usedFeatures.size()]=(int)SourceData.get(i)[MaxColumns-1];
			DestinationData.add(tmpArr);
		}

	}


	/**Prepares the NonMarInCurrentDataset data member.
	 */	

	private void findNonMarInCurrentDataset(){

		nonMarInCurrentTrainingDataset= new ArrayList<Integer>(nonMarFeatureNamesTraining.size()) ;
		nonMarInCurrentTestingDataset= new ArrayList<Integer>(nonMarFeatureNamesTesting.size()) ;
		numClassesNonMarTesting= new ArrayList<Integer>(nonMarFeatureNamesTesting.size()) ;
		int IdxFullDataset=-1;
		int i;

		for (String tmpStr : nonMarFeatureNamesTraining)
		{
			//==get the index, unfortunately the array is unsorted
			for (i=0; i<usedFeaturesNames.size(); i++)
			{
				if (usedFeaturesNames.get(i).equalsIgnoreCase(tmpStr))
				{
					nonMarInCurrentTrainingDataset.add(i);
					break;
				}
			}
		}


		for (String tmpStr : nonMarFeatureNamesTesting)
		{
			//==get the index, unfortunately the array is unsorted
			for (i=0; i<usedFeaturesNames.size(); i++)
			{
				if (usedFeaturesNames.get(i).equalsIgnoreCase(tmpStr))
				{
					nonMarInCurrentTestingDataset.add(i);
					break;
				}
			}

			//we have to search by hand, as the array is not sorted!	
			//numerical feature

			for (i=0; i<featNames.size(); i++)
			{
				if (featNames.get(i).equalsIgnoreCase(tmpStr))
				{
					IdxFullDataset=i;
					break;
				}
			}

			if (numFlags.get(IdxFullDataset))
			{
				if (discretizationIntervals[IdxFullDataset][0]==Double.MAX_VALUE)
					continue;
				numClassesNonMarTesting.add((discretizationIntervals[IdxFullDataset].length)+1);
			}
			//categorical feature
			else
				numClassesNonMarTesting.add(categoryNames.get(IdxFullDataset).length);	
		}
	}



	/** 
	 * Discretizes all the numerical features on the Training Set, and instantiates DiscretizationIntervals,
	 * UsedFeatures, UsedFeaturesNames, NumClassForEachUsedFeature; updates DiscretizationLog.
	 */
	private void discretizeNumFeaturesOnTrainingData (ArrayList<double[]> TrainingData)
	{
		discretizationIntervals=new double[numFlags.size()][];
		//avoid effects of previous CV-runs
		MdlDiscretizer Discretizer;

		usedFeatures=new ArrayList<Integer>(featNames.size());
		notUsedFeatures=new ArrayList<Integer>(featNames.size());

		//System.out.println("\n\n**Discretization of numerical features:");
		for (int j = 0; j < numFlags.size(); j++)
		{
			if (numFlags.get(j)) 
			{
				ArrayList<Integer> ClassValues=new ArrayList<Integer>(TrainingData.size());
				ArrayList<Double> FeatureValues = new ArrayList<Double>(TrainingData.size());
				for (int k = 0; k < TrainingData.size(); k++) {
					if (TrainingData.get(k)[j] != -9999) {
						ClassValues.add((int)TrainingData.get(k)[numFlags.size()]);
						FeatureValues.add(TrainingData.get(k)[j]);
					}
				}

				int NonDiscretizedCounter=0;
				Discretizer= new MdlDiscretizer(FeatureValues, ClassValues,classesNames.size());

				if (Discretizer.getAcceptedCutPoints().size()>0)
					discretizationIntervals[j]=ArrayUtils.arrList2Array(Discretizer.getAcceptedCutPoints());
				else
					discretizationIntervals[j]=new double[]{Double.MAX_VALUE};

				if (Discretizer.getAcceptedCutPoints().size()>0) 
					usedFeatures.add(j);
				else 
				{
					notUsedFeatures.add(j);
					NonDiscretizedCounter++;
				}
			}//end if numerical

			else 
			{ 
				//remove categorical features which have a single category only
				if (categoryNames.get(j).length >1)
					usedFeatures.add(j);
				else 
					notUsedFeatures.add(j);

			}
		}//end ==for on features

		Collections.sort(usedFeatures);
		Collections.sort(notUsedFeatures);
		usedFeaturesNames=new ArrayList<String>(usedFeatures.size());
		numClassForEachUsedFeature=new ArrayList<Integer>(usedFeatures.size());
		int i;
		for (i=0; i<usedFeatures.size(); i++)
		{
			usedFeaturesNames.add(i, featNames.get(usedFeatures.get(i)));
			if (numFlags.get(usedFeatures.get(i)))
				numClassForEachUsedFeature.add(i,(discretizationIntervals[usedFeatures.get(i)].length)+1);
			else
				numClassForEachUsedFeature.add(i,categoryNames.get(usedFeatures.get(i)).length);
		}

		findNonMarInCurrentDataset();

		for (int z=0; z<notUsedFeatures.size(); z++)
			//categorical feature with a single category don't go here;
			//we look for numerical feat discretized into a single bin
			if (numFlags.get(notUsedFeatures.get(z)))
				discretLog[notUsedFeatures.get(z)]++;
	}




//	=====PRIVATE HELPER FUNCTIONS=====




	/**Given a numerical value of a certain discretized feature, 
	 * returns the index of the bin in which the value falls
	 */
	private int getDiscretizationIdx(Double currentValue,
			int FeatureIdx) {

		int NextIdx=1;
		int PreviousIdx=0;
		int NumCutPoints=discretizationIntervals[FeatureIdx].length;

		if (currentValue<discretizationIntervals[FeatureIdx][0])
			return 0;

		if (currentValue>discretizationIntervals[FeatureIdx][NumCutPoints-1])
			return NumCutPoints;

		//on testing data, it may happen that a datum is actually equal to a cutpoint
		int MatchingCutPoint=Arrays.binarySearch(discretizationIntervals[FeatureIdx], currentValue);

		if (MatchingCutPoint < 0)
		{
			//we have an intermediate value
			while (! (currentValue<discretizationIntervals[FeatureIdx][NextIdx]) &
					(currentValue>discretizationIntervals[FeatureIdx][PreviousIdx]) )
			{
				PreviousIdx=NextIdx++;
			}
			return 	NextIdx;
		}
		else
			return MatchingCutPoint;
	}


//	=====DATA MEMBERS=====

	/**
	 * Dataset Name as read from the field "@relation" in the Arff file
	 */
//	private String DatasetName;



	/** Set either to "CV" or to the name of the testing Arff file */
	private String validationMethod;


	/**
	 * Names of  NonMar features in training
	 */
	private ArrayList<String> nonMarFeatureNamesTraining;

	/**
	 * Names of  NonMar features in testing
	 */
	private ArrayList<String> nonMarFeatureNamesTesting;

	/**
	 * Object for parsing ARFF files
	 */
	private ArffParser aParser;

	/**
	 * Index of NonMar features positions in the current 
	 * training set (position might change during CV, as different variables can get discretized into a single bin)
	 */
	private ArrayList<Integer> nonMarInCurrentTrainingDataset;

	/**
	 * Index of NonMar features positions in the current 
	 * testing set (position might change during CV, as different variables can get discretized into a single bin)
	 */	private ArrayList<Integer> nonMarInCurrentTestingDataset;

	 /**Number of classes of variables NonMar in the testing set. 
	  * Useful when the NCC builds all the possible realizations of 
	  * the NonMar variables
	  */
	 private ArrayList<Integer>  numClassesNonMarTesting;

	 /** Names of input features */
	 private ArrayList<String> featNames;

	 /** Flags array, regarding wheter Features are numerical (1) or not (0)*/
	 private ArrayList<Boolean> numFlags;

	 /**Variables used in the current experiment, hence excluding those
	 discretized in a single bin. Indexes refer to RawDataset
	  * */
	 private ArrayList<Integer> usedFeatures;

	 /**Names of the variables used in the current experiment, hence excluding those
	 discretized in a single bin.
	  * */
	 private ArrayList<String> usedFeaturesNames;

	 /**
	  *Number of classes for each used feature  
	  */
	 private ArrayList<Integer> numClassForEachUsedFeature;

	 /**Variables not used in the current experiment,
	   because discretized in a single bin;
	   indexes refer to RawDataset
	  * */
	 private ArrayList<Integer> notUsedFeatures;


	 /**
	  * Copy of the data read from Arff file (having hence -9999 as marker for
	  * missing data), and category names substituted by the corresponding
	  * indexes.)
	  */
	 private ArrayList<double[]> rawDataset;

	 /**
	  * Raw testing set exactly as read from file.
	  * Used when a testing file with unknown classes is provided, to eventually 
	  * dump to file the values of the instances.
	  * Being declared as String[][], it  hosts number as well as
	  * categories.
	  */
	 private ArrayList<String[]> rawTestingSet;


	 /**
	  * Matrix with rows of different length; stores the bin ranges for numerical
	  * features
	  */
	 private double[][] discretizationIntervals;



	 /**
	  * How many times each feature has been discretized in a single bin,
	  * over the different training/testing experiments.
	  */
	 private int[] discretLog;




	 /**
	  * Matrix of String with rows of different lenght, as different features
	  * (each row of the matrix corresponds to a different feature) can have
	  * different numbers of categories.
	  * <p>
	  */
	 private ArrayList<String[]> categoryNames;


	 /**
	  * File that reports avg and std dev of performance indicators; this is the ultimate output file
	  */
	 private String resFile;

	 /**
	  * Path where the files for the given experiment (Arff files, NonMar.txt) are found,
	  * and where output files will be saved.
	  */
	 private String workPath;


	 /** Absolute Path of the testing Arff file */
	 private String arffTestingFile;
	 /** Absolute Path of the predictions file for CV*/
	 private String predsFile;

	 /** Name of the testing Arff file */
	 private String arffTestingFileAddress;





	 /**
	  * Names of the output class.
	  */
	 private ArrayList<String> classesNames;

	 /**
	  * Indexes of the rows, in RawDataset, which have the same output class.
	  * For instance, the first row collects the indexes of all the rows in RawDataset
	  * having output
	  * class c1, and so on.
	  */
	 private ArrayList<Integer>[] rowsClassIdx;

	 /**
	  * Indexes for cross validation: in which fold each row of 
	  * RawDataset falls
	  */
	 private int[] cvFoldsIdx;

	 /**
	  * How many instances are in each fold
	  */
	 private int[] foldsSize;

	 /** Number of folds used by cross-validation */
	 private int numCvFolds;

	 /** Number of Cross validation Runs*/
	 private int numCrossVRuns;

	 /**
	  * Training set, accessed by the classifier: numerical
	  * variables are discretized, while category names and classes are substituted by 
	  * indexes; missing data denoted as -9999.
	  */
	 private ArrayList<int[]> trainingSet;

	 /**
	  * Testing set, accessed by the classifier: numerical
	  * variables are discretized, while category names and classes are substituted by 
	  * indexes; missing data denoted as -9999.
	  */
	 private ArrayList<int[]> testingSet;


	 //==NESTED CLASS: ResultsReporter
	 /**
	  * Helper class for jncc, which accomplishes the following tasks: 
	  * reads the temporary file where NBC and NCC predictions are stored; 
	  * computes performances indexes; produces the output file, which reports both
	  * the discretization log (i.e., whether some numerical feats have been discretized into a single bin)
	  * and the classifiers results.
	  */
	 private static class ResultsReporter {

		 /**
		  * Constructor
		  */
		 ResultsReporter(int[] SuppliedDiscretizationLog, String suppliedResultsFile, ArrayList<String> suppliedFeatureNames,
				 ArrayList<String> suppliedClassNames, String suppliedPredictionsFile, String suppliedWorkingPath, 
				 int suppliedNumFolds, int suppliedNumCVRuns)
				 {
			 discretizationLog=SuppliedDiscretizationLog;
			 resultsFile=suppliedResultsFile;
			 featureNames=suppliedFeatureNames;
			 classNames=suppliedClassNames;
			 predictionsFile=suppliedPredictionsFile;
			 workingPath=suppliedWorkingPath;
			 numFolds=suppliedNumFolds;
			 numCVRuns=suppliedNumCVRuns;
			 numRuns=numFolds*numCVRuns;
			 //testing set 
			 if (numRuns==0)
				 numRuns=1;
				 }


		 /**
		  * Writes to the output file how many times the different variables have been discretized
		  * into a single bin, out of the total number of performed experiments. 
		  */
		 void writeDiscretizationLog(int totalExp)
		 {
			 try{
				 BufferedWriter out=new BufferedWriter(new FileWriter(resultsFile,false));

				 out.write("Validation Method: ");
				 if (totalExp>1)
					 out.write("\n Cross-validation\n");
				 else
					 out.write ("Testing File");

				 for (int i=0; i<discretizationLog.length; i++)
				 {
					 if (discretizationLog[i]>0)
						 out.write("\nFeature "+featureNames.get(i)+
								 " discretized into a unique bin in "+discretizationLog[i]+"/"+totalExp+
						 " induction experiments");
				 }
				 out.close();
			 }

			 catch (IOException e)
			 {
				 System.out.println("Problems writing discretization log");
			 }
		 }


		 /**
		  * Parses the (temporary) file where the performance indicator have been stored fold
		  * by fold, and outputs to file the report. Standard deviation is empirically computed 
		  * between the folds.
		  */
		 void writeStatsToFile()
		 {
			 try
			 {
				 //skip first line, containing  comments
				 LineNumberReader lnr = new LineNumberReader(new InputStreamReader(
						 new FileInputStream(indicatorsFile)));
				 String line = "";

				 double[] NBCacc = new double[numRuns];
				 double[][] PerClassNBCacc = new double[classNames.size()][numRuns];
				 double[] NCCprec = new double[numRuns];
				 double[] NCCsingleAcc = new double[numRuns];
				 double[] NCCsetAcc = new double[numRuns];
				 double[] NCCoutputSize = new double[numRuns];
				 double[] NCCP = new double[numRuns];
				 double[] NCCI = new double[numRuns];

				 int i = -1;

				 //skip first line, containing indicators names
				 lnr.readLine();
				 while ((line = lnr.readLine()) != null) {

					 i++;
					 //parse data
					 StringTokenizer st = new StringTokenizer(line, "\t");
					 Double tmpDouble;
					 NBCacc[i]=Double.parseDouble(st.nextToken());

					 for (int j=0; j<classNames.size();j++)
					 {
						 tmpDouble=Double.parseDouble(st.nextToken());
						 if (tmpDouble.isNaN())
							 PerClassNBCacc[j][i]=-9999;
						 else
							 PerClassNBCacc[j][i]=tmpDouble;
					 }
					 NCCprec[i]=Double.parseDouble(st.nextToken());

					 tmpDouble=Double.parseDouble(st.nextToken());
					 if (tmpDouble.isNaN())
						 NCCsingleAcc[i]=-9999;
					 else
						 NCCsingleAcc[i]=tmpDouble;


					 tmpDouble=Double.parseDouble(st.nextToken());
					 if (tmpDouble.isNaN())
						 NCCsetAcc[i]=-9999;
					 else
						 NCCsetAcc[i]=tmpDouble;


					 tmpDouble=Double.parseDouble(st.nextToken());
					 if (tmpDouble.isNaN())
						 NCCoutputSize[i]=-9999;
					 else
						 NCCoutputSize[i]=tmpDouble;

					 tmpDouble=Double.parseDouble(st.nextToken());
					 if (tmpDouble.isNaN())
						 NCCP[i]=-9999;
					 else
						 NCCP[i]=tmpDouble;


					 tmpDouble=Double.parseDouble(st.nextToken());
					 if (tmpDouble.isNaN())
						 NCCI[i]=-9999;
					 else
						 NCCI[i]=tmpDouble;

				 }
				 lnr.close();
				 //==INDICATORS FILE HAS BEEN PARSED 



				 BufferedWriter out=new BufferedWriter(new FileWriter(resultsFile,true));
				 int j=0;
				 DecimalFormat formatter= new DecimalFormat("#0.00%");
				 //==WRITING NBC STATS
				 out.write("\n\n----------------------\n");
				 out.write("Naive Bayesian Classifier\n\n");
				 out.write("\nACCURACY\n");
				 out.write(formatter.format(ArrayUtils.arrayAvg(NBCacc)));
				 if (numCVRuns>0)
					 out.write(" +- "+ formatter.format(ArrayUtils.arrayStDev(NBCacc)));

				 out.write("\n");

				 //==per-class accuracy
				 out.write("\n==PER-CLASS ACCURACY\n");
				 for (i=0; i<classNames.size(); i++)
				 {
					 if (ArrayUtils.arrayAvg(PerClassNBCacc[i])==-9999)
						 //there are no samples for this class in the testing set
						 out.write ("\n"+classNames.get(i)+":  never present in testing set");
					 else
					 {
						 out.write ("\n"+classNames.get(i)+":  "+ formatter.format(ArrayUtils.arrayAvg(PerClassNBCacc[i])));
						 if (numCVRuns>0){
							 out.write(" +- "+ formatter.format(ArrayUtils.arrayStDev(PerClassNBCacc[i])));
						 }
					 }
				 }

				 //==CONFUSION MATRIX
				 out.write("\n\n==CONFUSION MATRIX\n\n");
				 for (i=0; i<classNames.size(); i++)
					 out.write(classNames.get(i)+"\t");
				 out.write("\t <--classified as \n");

				 //procedure for cross validation
				 if (numCVRuns>0)
				 {	
					 DecimalFormat formatter3= new DecimalFormat("#0");
					 for (i=0; i<nbcConfusionMatrix.length; i++)
					 {
						 for (j=0; j<nbcConfusionMatrix[i].length-1; j++)
							 out.write(formatter3.format(Math.round((double)(nbcConfusionMatrix[i][j]/numCVRuns)))+"\t"); 
						 out.write(formatter3.format (Math.round((double)(nbcConfusionMatrix[i][j]/numCVRuns)))
								 +"\t"+classNames.get(i)+"\n");
					 }
				 }

				 //Testing File
				 else
				 {
					 for (i=0; i<nbcConfusionMatrix.length; i++)
					 {
						 for (j=0; j<nbcConfusionMatrix[i].length-1; j++)
							 out.write(nbcConfusionMatrix[i][j]+"\t");
						 out.write(nbcConfusionMatrix[i][j]+"\t"+classNames.get(i)+"\n");
					 }
				 }

				 //==NBC STATS HAVE BEEN WRITTEN



				 //==WRITING NCC STATS FROM HERE ON
				 out.write("\n\n\n----------------------");
				 out.write("\nNaive Credal Classifier2\n");


				 //standard deviation printed only for cross validation

				 out.write("\n\n==PRECISE CLASSIFICATIONS\n"+ formatter.format(ArrayUtils.arrayAvg(NCCprec)));
				 if (numCVRuns>0)
					 out.write(" +- "+ formatter.format(ArrayUtils.arrayStDev(NCCprec))+"\n");


				 if (ArrayUtils.arrayAvg(NCCsingleAcc)==-9999)
					 out.write("\nNBC-single accuracy not defined\n\n");
				 else
				 {
					 out.write("\n==SINGLE-ACCURACY\n"+ formatter.format(ArrayUtils.arrayAvg(NCCsingleAcc)));
					 if (numCVRuns>0)
						 out.write(" +- "+formatter.format(ArrayUtils.arrayStDev(NCCsingleAcc))+"\n");
				 }

				 if (ArrayUtils.arrayAvg(NCCsetAcc)==-9999)
					 out.write("Set-accuracy and imprecise output size not defined\n");
				 else{
					 out.write("\n==SET-ACCURACY\n"+ formatter.format(ArrayUtils.arrayAvg(NCCsetAcc)));
					 if (numCVRuns>0)
						 out.write(" +- "+formatter.format(ArrayUtils.arrayStDev(NCCsetAcc)));
					 DecimalFormat intFormatter= new DecimalFormat("#0.0");
					 out.write("\n\n==AVERAGE SIZE OF IMPRECISE OUTPUT\n "+ intFormatter.format(ArrayUtils.arrayAvg(NCCoutputSize)));	
					 if (numCVRuns>0)
						 out.write(" +- "+intFormatter.format(ArrayUtils.arrayStDev(NCCoutputSize)));
				 }

				 //==CONFUSION MATRIX
				 out.write("\n\n== NCC2 CONFUSION MATRIX (computed on precisely classified instances only)\n\n");
				 for (i=0; i<classNames.size(); i++)
					 out.write(classNames.get(i)+"\t");
				 out.write("\t <--classified as \n");

				 if (numCVRuns>0)
				 {	
					 DecimalFormat formatter3= new DecimalFormat("#0");
					 for (i=0; i<nccConfusionMatrix.length; i++)
					 {
						 for (j=0; j<nccConfusionMatrix[i].length-1; j++)
							 out.write(formatter3.format(Math.round((double)(nccConfusionMatrix[i][j]/numCVRuns)))+"\t"); 
						 out.write(formatter3.format (Math.round((double)(nccConfusionMatrix[i][j]/numCVRuns)))
								 +"\t"+classNames.get(i)+"\n");
					 }
				 }

				 //Testing File
				 else
				 {
					 for (i=0; i<nccConfusionMatrix.length; i++)
					 {
						 for (j=0; j<nccConfusionMatrix[i].length-1; j++)
							 out.write(nccConfusionMatrix[i][j]+"\t");
						 out.write(nccConfusionMatrix[i][j]+"\t"+classNames.get(i)+"\n");
					 }
				 }
				 //==END CONFUSION MATRIX

				 out.write("\n\n\n----------------------");
				 out.write("\nAnalysis of NBC accuracy on subsets of instances\n");
				 if (ArrayUtils.arrayAvg(NCCsingleAcc)==-9999) 
					 out.write("\nNCC2 always imprecise!\n");
				 else if (ArrayUtils.arrayAvg(NCCsetAcc)==-9999)
					 out.write("\nNCC2 always precise!\n");
				 else
				 {
					 out.write("\nNBC accuracy when NCC2 is precise:" +
							 "\t"+ formatter.format(ArrayUtils.arrayAvg(NCCP)));
					 if (numCVRuns>0)
						 out.write(" +- " + formatter.format(ArrayUtils.arrayStDev(NCCP)));

					 out.write("\nNBC accuracy  when NCC2 is imprecise:" +
							 "\t"+ formatter.format(ArrayUtils.arrayAvg(NCCI)));
					 if (numCVRuns>0)
						 out.write(" +- " + formatter.format(ArrayUtils.arrayStDev(NCCI)));				
				 }

				 out.close();

				 //delete predictions file
				 boolean success = (new File(predictionsFile)).delete();
				 if (!success) {
					 // Deletion failed
					 System.out.println("Impossible deleting file "+predictionsFile+" from filesystem.");
				 }

				 //delete indicators file
				 success = (new File(indicatorsFile)).delete();
				 if (!success) {
					 // Deletion failed
					 System.out.println("Impossible deleting file "+indicatorsFile+" from filesystem.");
				 }

			 }


			 catch (IOException ioexc) {
				 System.out.println("Unexpected exception reading file " + predictionsFile+
				 " or writing results to file. Please check directory permission. ");
				 System.exit(0);
			 } 
		 }



		 /**
		  *Computes statistics accuracy of NBC and NCC, by analyzing the predictions saved on a temporary file; then,
		  *saves the file the computed indicators and deletes the temporary file.
		  *The file is expected to be made up by many rows arranged as follows:<p>
		  *FoldNumber | FEATURES(m columns) | Actual Class (1 col) | NBC prediction (1 col) | NCC prediction (nc cols)
		  *<p>
		  *where m is the number of features of the data set, and nc is the number of classes.
		  *The function properly manages the fact that in different runs of CV the number of features can
		  *be different, because of the discretization (i.e., in some runs, some features might be discretized into a unique bin).
		  */
		 void analyzePredictionsFileNbcNcc ()
		 {

			 try {




				 //DECLARATIONS AND INITIALIZATIONS
				 LineNumberReader lnr = new LineNumberReader(new InputStreamReader(
						 new FileInputStream(predictionsFile)));
				 String line = "";
				 String currentToken = "";
				 int ActualClass=0;
				 int NCCImprecise=0;
				 int NCCPrecise=0;
				 int NCCPreciseAccurate=0;
				 int NCCSetAccurate=0;
				 int NCCImpreciseOutputSize=0;
				 int NBCAccNCCImprecise=0;
				 int NBCAccNCCprecise=0;

				 //tmp file where to store the indicators (a row of indicators for each fold)
				 indicatorsFile =workingPath+"indicators.csv";
				 BufferedWriter out2=new BufferedWriter(new FileWriter(indicatorsFile,false));
				 out2.write("NBCacc\t"); 
				 for (int i=0; i<classNames.size(); i++)
					 out2.write("NBCaccC"+i+"\t"); 
				 out2.write("NCCprec\tNCCacc\tNCCsacc\tNCCoutsize\tNCCP\tNCCI\n");
				 out2.close();
				 out2=new BufferedWriter(new FileWriter(indicatorsFile,true));




				 int NBCPrediction;
				 boolean NBCAccurate;
				 int NumClasses=classNames.size();
				 int InstancesCounter=0;
				 int[] NbcPerClassAccurate= new int [NumClasses];
				 Arrays.fill(NbcPerClassAccurate,0);
				 int[] ClassFrequencies= new int [NumClasses];
				 Arrays.fill(ClassFrequencies,0);
				 nbcConfusionMatrix= new int [NumClasses][NumClasses];
				 nccConfusionMatrix= new int [NumClasses][NumClasses];
				 int i;
				 for (i=0;i<nbcConfusionMatrix.length;i++)	
				 {
					 Arrays.fill(nbcConfusionMatrix[i], 0);
					 Arrays.fill(nccConfusionMatrix[i], 0);
				 }
				 int currentFold= -9999;
				 int previousFold= -9999;
				 //END DECLARATIONS AND INITIALIZATIONS



				 //==PARSE PREDICTIONS IN FILE
				 while ((line = lnr.readLine()) != null) {


					 //parse data
					 StringTokenizer st = new StringTokenizer(line, "\t");


					 //NumFeatures is allowed to change betwen different blocks of CV because of
					 //the discretization.
					 int NumFeatures=st.countTokens()-3-NumClasses;
					 int MissingInCurrentInstance=0;

					 currentFold=Integer.parseInt(st.nextToken());
					 if (previousFold==-9999)
						 previousFold=currentFold;

					 //folder is changed, save to file the stats on the current fold 
					 //and set to zero the various counters
					 if (!(currentFold==previousFold)) 
					 {
						 //NBC accuracy
						 out2.write(ArrayUtils.arraySum(NbcPerClassAccurate)[0]/(double)InstancesCounter+"\t");
						 //NBC per-class accuracy
						 for (i=0; i<classNames.size();i++)
							 out2.write ((double)NbcPerClassAccurate[i]/(double)ClassFrequencies[i]+"\t");

						 //NCC precision
						 out2.write ((double)NCCPrecise/(double)InstancesCounter+"\t");
						 //NCC single acc
						 out2.write ((double)NCCPreciseAccurate/(double)NCCPrecise+"\t");
						 //NCC set acc
						 out2.write ((double)NCCSetAccurate/((double)InstancesCounter-(double)NCCPrecise)+"\t");
						 //NCC out size
						 out2.write((double)NCCImpreciseOutputSize/((double)InstancesCounter-(double)NCCPrecise)+"\t");
						 //NCCP
						 out2.write((double)NBCAccNCCprecise/(double)NCCPrecise+"\t");
						 //NCCI
						 out2.write((double)NBCAccNCCImprecise/(double)NCCImprecise+"\n");



						 //reset counters relevant to current fold
						 NCCImprecise=0;
						 NCCPrecise=0;
						 NCCPreciseAccurate=0;
						 NCCSetAccurate=0;
						 NCCImpreciseOutputSize=0;
						 NBCAccNCCImprecise=0;
						 NBCAccNCCprecise=0;
						 Arrays.fill(ClassFrequencies,0);
						 InstancesCounter=0;
						 NbcPerClassAccurate= new int [NumClasses];
						 previousFold=currentFold;
					 }


					 for (i =0; i<NumFeatures; i++)
					 {
						 currentToken = st.nextToken();
						 if (currentToken.compareTo("-9999")==0)
							 MissingInCurrentInstance++;
					 }
					 ActualClass=Integer.parseInt(st.nextToken());
					 ClassFrequencies[ActualClass]++;

					 //parsing NCC prediction
					 int CurrentOutputSize=0;
					 Integer NCCPredictedClass=new Integer(0);
					 int NCCFirstPrediction=new Integer(0);
					 boolean NCCAccurate=false;
					 for (i =0; i<NumClasses; i++)
					 {
						 currentToken=st.nextToken();
						 NCCPredictedClass = new Integer(currentToken);
						 if (i==0)
							 NCCFirstPrediction=new Integer(currentToken);
						 if (NCCPredictedClass !=6666)
							 CurrentOutputSize++;
						 if (NCCPredictedClass.compareTo(ActualClass)==0)
							 NCCAccurate=true;
					 }

					 //parsing NBC prediction
					 currentToken=st.nextToken();
					 NBCPrediction=Integer.parseInt(currentToken);
					 NBCAccurate=(NBCPrediction==ActualClass);
					 if (NBCAccurate)
						 NbcPerClassAccurate[ActualClass]++;
					 nbcConfusionMatrix[ActualClass][NBCPrediction]++;


					 //updating indicators
					 if (CurrentOutputSize>1)
					 {
						 NCCImprecise++;
						 NCCImpreciseOutputSize+=CurrentOutputSize;
						 if (NBCAccurate)
							 NBCAccNCCImprecise++;
						 if (NCCAccurate)
							 NCCSetAccurate++;
					 }

					 else

					 {
						 NCCPrecise++;
						 if (NBCAccurate)
							 NBCAccNCCprecise++;
						 if (NCCAccurate)
							 NCCPreciseAccurate++;
						 nccConfusionMatrix[ActualClass][NCCFirstPrediction]++;
					 }
					 InstancesCounter++;
				 }


				 //==DUMP LAST FOLD RESULTS
				 //NBC accuracy
				 out2.write(ArrayUtils.arraySum(NbcPerClassAccurate)[0]/(double)InstancesCounter+"\t");
				 //NBC per-class accuracy
				 for (i=0; i<classNames.size();i++)
					 out2.write ((double)NbcPerClassAccurate[i]/(double)ClassFrequencies[i]+"\t");
				 //NCC precision
				 out2.write ((double)NCCPrecise/(double)InstancesCounter+"\t");
				 //NCC single acc
				 out2.write ((double)NCCPreciseAccurate/(double)NCCPrecise+"\t");
				 //NCC set acc
				 out2.write ((double)NCCSetAccurate/((double)InstancesCounter-(double)NCCPrecise)+"\t");
				 //NCC out size
				 out2.write((double)NCCImpreciseOutputSize/((double)InstancesCounter-(double)NCCPrecise)+"\t");
				 //NCCP
				 out2.write((double)NBCAccNCCprecise/(double)NCCPrecise+"\t");
				 //NCCI
				 out2.write((double)NBCAccNCCImprecise/(double)NCCImprecise+"\n");

				 //==PREDICTIONS HAVE BEEN ANALYZED
				 out2.close();

			 }


			 catch (FileNotFoundException exc) {
				 System.out.println("Missing File " + predictionsFile);
				 System.out.println("Please check path and file name!");
				 System.exit(0);
			 } catch (IOException ioexc) {
				 System.out.println("Unexpected exception reading file " + predictionsFile+
				 " or writing results to file. Please check directory permission. ");
				 System.exit(0);
			 } 


		 }



		 /**
		  * NBCconfusionMatrix, as parsed from the prediction file
		  */
		 private int[][]nbcConfusionMatrix;


		 /**
		  * NCCconfusionMatrix, as parsed from the prediction file
		  */
		 private int[][]nccConfusionMatrix;


		 /**
		  * How many times each feature has been discretized in a single bin,
		  * over the different training/testing experiments.
		  */
		 private int[] discretizationLog;

		 private String resultsFile;

		 /** Names of input features */
		 private ArrayList<String> featureNames;

		 /**
		  * Temporary file, where the performance indicators are saved fold-by-fold 
		  */
		 private String indicatorsFile;

		 /**
		  * how many training/testing experiments (1 for testing, num runs*num folds for CV) the predictions saved to file reg
		  */
		 private int numRuns;

		 private int numCVRuns;
		 private int numFolds;


		 /**
		  * Names of the output class.
		  */
		 private ArrayList<String> classNames;

		 /**
		  * Where to look for predictions saved to file
		  */
		 private String predictionsFile;

		 private String workingPath;
	 }
}











